/*
 *  md5crypt - MD5 based authentication routines
 */
#include <stdio.h>
#include "ntp_types.h"
#include "ntp_fp.h"
#include "ntp_string.h"
#include "md5.h"
#include "ntp_stdlib.h"

extern u_long cache_keyid;
extern char *cache_key;
extern int cache_keylen;

#ifndef HAVE_MEMMOVE
extern void *memmove P((void *, const void *, size_t));
#endif

/*
 * Stat counters, imported from data base module
 */
extern u_long authencryptions;
extern u_long authdecryptions;
extern u_long authkeyuncached;
extern u_long authnokey;

extern int debug;

/*
 * For our purposes an NTP packet looks like:
 *
 *	a variable amount of encrypted data, multiple of 8 bytes, followed by:
 *	NOCRYPT_OCTETS worth of unencrypted data, followed by:
 *	BLOCK_OCTETS worth of ciphered checksum.
 */ 
#define	NOCRYPT_OCTETS	4
#define	BLOCK_OCTETS	16

#define NTP_MAXKEY	65535	/* max nonauto key identifier from ntp.h */

int
MD5authencrypt(
	u_long keyno,
	u_int32 *pkt,
	int length		/* length of encrypted portion of packet */
	)
{
	MD5_CTX ctx;
	int len;		/* in 4 byte quantities */
	u_int32 keyid;

	authencryptions++;

	if (keyno != cache_keyid) {
		authkeyuncached++;
		if (!authhavekey(keyno)) {
			authnokey++;
			return 0;
		}
	}

	len = length / sizeof(u_int32);
	keyid = (u_int32)ntohl(keyno);

	/*
	 *  Generate the authenticator.
	 */
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned const char *)cache_key, cache_keylen);
	MD5Update(&ctx, (unsigned const char *)pkt, length);
	MD5Final(&ctx);

	memmove((char *)&pkt[len], &keyid, sizeof(u_int32));
	memmove((char *)&pkt[len + 1], (char *)ctx.digest, BLOCK_OCTETS);
	return (4 + BLOCK_OCTETS);	/* return size of key and MAC  */
}

/*
 * Generate session key from supplied plaintext.
 */
u_long
session_key(
	u_int32 srcadr,		/* source address */
	u_int32 dstadr,		/* destination address */
	u_long keyno,		/* key identifier */
	u_long lifetime		/* key lifetime */
	)
{
	MD5_CTX ctx;
	u_int32 header[3];
	u_long keyid;

	/*
	 * Generate the session key and retrieve the hash for later. If
	 * the lifetime is greater than zero, call the key trusted.
	 */
	header[0] = htonl(srcadr);
	header[1] = htonl(dstadr);
	header[2] = htonl(keyno);
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned const char *)header, sizeof(header));
	MD5Final(&ctx);
	memcpy(&keyid, ctx.digest, sizeof(u_int32));
	if (lifetime != 0) {
		MD5auth_setkey(keyno, (const char *)ctx.digest, sizeof(ctx.digest));
		authtrust(keyno, lifetime);
	}
#ifdef DEBUG
	if (debug > 1)
		printf(
		    "session_key: from %s to %s keyid %08lx hash %08lx life %d\n",
		    numtoa(htonl(srcadr)), numtoa(htonl(dstadr)), keyno,
		    keyid, lifetime);
#endif
	return (keyid);
} 
